// KEDS_count  Version 3.0b1 [Java]

// Kansas Event Data System Project
// Philip A. Schrodt
// Dept of Politial Science
// University of Kansas
// Lawrence, KS 66045
// schrodt@ku.edu

// Revision History
// 1.0 Philip Schrodt in Pascal 1995
// 2.0 Star Straf rewrite in Java Jan-2000
// 2.1 Modifications by Schrodt March-2007
// 3.0 Modifications by Schrodt June-2007


// To compile: javac KEDS_count.java
// To run: java KEDS_count

//__________________________________________________________________________________
//
// 	Redistribution and use in source and binary forms, with or without modification,
// 	are permitted under the terms of the GNU General Public License:
// 	http://www.opensource.org/licenses/gpl-license.html
//
//	Report bugs to: schrodt@ku.edu

//	The most recent version of this code is available from the KEDS Web site:
//		http://web.ku.edu/keds
//__________________________________________________________________________________
//


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Calendar;
import java.text.*;


class KEDS_count extends Frame implements ActionListener
{
// CONSTANTS
static final String VER = "Version 3.0b1";
static final int CODEVALUEMAX = 1024;
static final int FILESMAX = 32;
static final int COUNTMAX = 64;
static final int TARGETSMAX = 32; // maximum targets in a DYADS statement
static final int ITEMSMAX   = 32; // maximum items in a [ ] source or target
static final String TAB = "\t";
static final String TEMPFILENAME = "KCount.temp";

// Command file variables divided by section
// - FORMAT  
  String 	formatPrefix 	= "KCOUNT.";
  int     formatDate 		= 1;
  int 		formatSource 	= 2;
  int 		formatTarget 	= 3;
  int 		formatEvent 	= 4;
  String 	formatFreq 		= "MONTHLY";
  String 	formatCalender 	= "yyMMdd";
  SimpleDateFormat formatSDF; 
 
  int			formatMark	= 250;    // count of how often to display =+*

// - SCALES  <code>:[value]
  String 	code[];
  float 	value[];
  int 		maxCodeValue 	= 0;

// - FILES
  String 		files[];
  int 			maxFiles = 0;

// - COUNT
  String 		count[][];     // identifiers
  boolean 	cWildcard[][]; // does identifer have a wildcard
  String 		cLabel[];
  int				cmaxItems[];   // number of identifiers in category
	int 			maxCount = 0;  // number of categories

// - DYADS
// Note <07.06.28> : it would have been more elegant to implement these as lists, but
//   								 the legacy code used arrays, so stick with arrays....
  String 		dSource[];
  boolean 	dWildcard[];
  int       maxSItems = 0;
  boolean 	tAll[];
  String 		dTarget[][];
  boolean 	dTWildcard[][];
  int				dTmaxItems[];
  String 		dLabel[];
	int 		  maxTargets = 0;

// Event
  String 		locDate;
  String 		locSrc;
  String 		locTar;
  String 		locEvt;

// Actual count for current Time element (day)
  int				dayCount[];
  float			dayScale[];
  int				dayCode[][];
  int  			firstWeekDay 	= Calendar.SUNDAY; // day of the week to start a week at
  Calendar	FirstDate;                     // initial date tabulated in data set
  Calendar	BeginDate, EndDate, CurrDate;  // bounds of current interval
  boolean		firstData = true;

// Display variables
  TextField 	status;
  TextArea 		dataDisplay;
  int					dspcnt, dspline;
  String			dspchar;
  
// files
	PrintWriter 	dyadOut[];  // individual dyad files
  PrintWriter 	composOut;   // composite output
  PrintWriter 	tempOut;    // temporary buffer for composite output
  BufferedReader composIn;   // composite input

// Output options
	String composFileName;
	boolean	doDyadFiles		= false;	// write results to individual files
	boolean	doComposFile	= false;		// write results to composite file
	boolean	hasCompos			= false; //  ComposFile has content


public void 	actionPerformed(ActionEvent event)
// Close window and exit program; required override function
{  
	Object source = event.getSource();
	System.exit(0);
}  // actionPerformed


void	KEDS_error(String error)
{
  this.dataDisplay.append("\nERROR encountered in program: " + error + "\n");
  System.out.println(error);
}

void	Fatal_error(String error)
{
  this.dataDisplay.append("\nFatal error encountered in program:\n" + error + "\n");
  System.out.println("\nFatal error encountered in program:\n" + error + "\n");
	System.exit(0);
}


// ******************************************************************
// *               Read in Command File                             *
// ******************************************************************

void 	KEDS_read_command_file()
// read in the command file and display progress on screen
{
  String 		thisLine 		= null;  // line read from line, then trimmed
  String 		inputFileName 	= "/Developer/KEDS_Count Java/TestData.KCJ.TAB";
  char 			section 		= ' ';
  FileDialog 	fdialog;
  
  this.dataDisplay.setText("Reading in command file");
  if (true) // ###  set to false for debugging
  {
		do {
			fdialog = new FileDialog(this, "Select Command File",FileDialog.LOAD);
			fdialog.setVisible(true);
			inputFileName = fdialog.getFile();
		} while (inputFileName == null);
		this.dataDisplay.setText("Command file: " + inputFileName + "\n");
	} // debugging
  try
  {
    File inputFile = new File(inputFileName);
    FileInputStream in = new FileInputStream(inputFile);
    BufferedReader br = new BufferedReader(new InputStreamReader(in));
    while ((thisLine = br.readLine()) != null)
    {  //  process the command file line by line
       thisLine = thisLine.trim();
       if (thisLine.startsWith("-")) { //  if a new section
				 this.dataDisplay.append("\n" + thisLine);
				 section = determine_section(thisLine);	   
	 		 } // new section if statement
			else if ((thisLine.length()>0) && (!thisLine.startsWith("#")))  // skip comments, blank lines
			{
				switch(section) {
					case 'F' : // -FORMAT
							this.format_section(thisLine);
							break;
					case 'S' : // -SCALES or -CODES
							this.scales_section(thisLine);
							break;  
					case 'L' : // -FILES
							this.files_section(thisLine);
							if (doComposFile) this.open_composite_file();
							break;  
					case 'C' : // -COUNT
							if (this.maxFiles == -1) { // no files in command file, prompt user for files
								this.manual_files_section();
							}
							this.parse_count(thisLine);
							break;  
					case 'D' : // -DYADS
							this.parse_dyad(thisLine);
							this.resetCounts();
							this.process_dyad();
							this.status.setText("STATUS: Processing DYADS");
							break;  
					case 'E' : // -FINISH
							break;  
					} // switch statement 
			}  //  data for a section (else)
    }  // while 
	  in.close();
  }  // try
  catch(java.io.IOException e) { KEDS_error("Cannot read from command file"); } 

}  //  KEDS_read_command_file

char determine_section(String secLine)
{  // determine which section you are at in the command file

  char temp;
  temp = ' ';

  if      (secLine.compareTo("-FORMAT") == 0)  temp = 'F';
  else if (secLine.compareTo("-SCALES") == 0)  temp = 'S';
  else if (secLine.compareTo("-CODES")  == 0)  temp = 'S'; // legacy from earlier version of documentation
  else if (secLine.compareTo("-FILES")  == 0)  temp = 'L';
  else if (secLine.compareTo("-COUNT")  == 0)  {
  	temp = 'C';
  	if (!doDyadFiles && !doComposFile) doComposFile = true; // set default
  }
  else if (secLine.compareTo("-DYADS")  == 0)  temp = 'D';
  else if (secLine.compareTo("-FINISH") == 0)  temp = 'E';
  else  this.KEDS_error( "Unknown section in command file" +	secLine);
  return temp;
}

void	format_section(String fmtLine)
{ // read in line from -FORMAT section of command file

  String tempCal;
  int pos = fmtLine.indexOf("=") + 1;

  if (fmtLine.startsWith("P"))  // PREFIX
	{ this.formatPrefix = fmtLine.substring(pos); }
  else if (fmtLine.startsWith("D"))  // DATE
	{ this.formatDate = Integer.parseInt(fmtLine.substring(pos)); }
  else if (fmtLine.startsWith("S"))  // SOURCE
 	{ this.formatSource =	Integer.parseInt(fmtLine.substring(pos)); }
  else if (fmtLine.startsWith("T"))  // TARGET
	{ this.formatTarget = 
		Integer.parseInt(fmtLine.substring(pos)); }
  else if (fmtLine.startsWith("E"))  // EVENT
 	{ this.formatEvent = Integer.parseInt(fmtLine.substring(pos)); }
  else if (fmtLine.startsWith("F")) { // FREQ
		this.formatFreq = fmtLine.substring(pos).trim(); 
		if (this.formatFreq.indexOf('(') > 0) { // shortcut here; just look for ( rather than (M)
		 	firstWeekDay = Calendar.MONDAY;
		 	this.formatFreq = this.formatFreq.substring(0,this.formatFreq.indexOf('('));
		}
	}
  else if ((fmtLine.startsWith("L") || (fmtLine.startsWith("C")))) { // LABEL, with CALENDAR as legacy
	  this.formatCalender = fmtLine.substring(pos);
 	  try { this.formatSDF = new SimpleDateFormat(this.formatCalender);}
 	  catch (IllegalArgumentException e) {
 	  		this.KEDS_error(" -LABEL format \'" + this.formatCalender + "\' is not a valid Java format string\n" 
 	  											+ "The default format \'yyMMdd\' will be used instead");
		}
  }
  else { this.KEDS_error("Unknown section in FORMAT section" + fmtLine); }
} // format_section

void scales_section(String sclLine)
 // read in <code>:[value] event codes and their scale
{
String 	cd,scs;
float  	scn;
//  int 		pos1,pos2,pos3;

int pos1 = sclLine.indexOf(":");
int pos2 = sclLine.indexOf("[");
int pos3 = sclLine.indexOf("]");

  if (pos1 == 0) {
		cd = null; 
		this.KEDS_error("Invalid code in SCALE section " + sclLine + " Ignored");
	}
  else {  cd = sclLine.substring(0,pos1); }

  if (pos2== 0) scn = 0;
  else { 
  	scs = sclLine.substring(pos2+1,pos3); 
	  scn = Float.valueOf(scs).floatValue();
	}

  if (this.maxCodeValue == CODEVALUEMAX-1) {
	  this.KEDS_error("Too many entries in SCALE section " + sclLine + " Ignored");
 	  cd = null;
	} 

  if (cd != null) {
	  this.code[this.maxCodeValue] = cd;
	  this.value[this.maxCodeValue] = scn;  
	  this.maxCodeValue++;
	}
} // scales_section

void files_section(String fileLine)
// process -FILE commands
{
  if (fileLine.indexOf("OUTPUT:") >= 0 ) {
  	if (fileLine.indexOf("COMP") >= 0 )  doComposFile = true; 
  	else if (fileLine.indexOf("INDIV") >= 0 ) doDyadFiles = true; 
  }
  else if (this.maxFiles == FILESMAX-1) {
	  this.KEDS_error("To many entries in FILE section " + fileLine + " Ignored");
 	}
  else {
	  if (fileLine.indexOf(":") > 0) {
	  	this.files[this.maxFiles] = fileLine.substring(fileLine.indexOf(":")+1).trim();}
	  else this.files[this.maxFiles] = fileLine.trim();    // 'INPUT:' is optional to handle legacy commands
	  this.maxFiles++;
	}
}	// files_section

void manual_files_section()
// if no files in command file then prompt user for files
// and store in KC.FileList.text
// <07.07.03>: this has not been tested
{
  FileDialog fdialog;
  String inputFileName = null;
  PrintWriter output;

  this.dataDisplay.append("Input data files \n");
  do {
		fdialog = new FileDialog(this,"Select Data File",FileDialog.LOAD);
		fdialog.setVisible(true);
		inputFileName = fdialog.getFile();
		if (inputFileName != null) {
			this.dataDisplay.append("Data file: " + inputFileName + "\n");
			this.files[this.maxFiles] = inputFileName;
			this.maxFiles++;
		}
  } while ((inputFileName != null) || (this.maxFiles < 0));

  try {
	  output = new PrintWriter(
		new FileOutputStream("KC.FileList.text"));
	}
  catch(java.io.IOException e2) {
	  this.KEDS_error("Problem opening KC.FileList.text");
	  return;
	}
  for ( int loopvar = 0; loopvar <= this.maxFiles; loopvar++) {
		output.println(this.files[loopvar]); }
  output.close();
  this.dataDisplay.append(
		"The file names have been written to the file" + " 'KC.FileList.text' \n" +
		"You can copy these into your command file for " +	"future use \n");
} //  manual_files_section

void parse_count(String cntLine)
{ // read in codes to be counted, may contain * ||
  
  int 		pos;  // end of current substring
  int			ka;   // start of current substring
  String 	subst;

  cntLine = cntLine.trim();
  cntLine += " ";
  pos = cntLine.indexOf(":");
  if (pos > 0)  { // get the label if it exists
  	cLabel[maxCount] = cntLine.substring(0,pos);
  	ka = pos + 1;
  }
  else {
  	cLabel[maxCount] = "";
  	ka = 0;
  }
  														// get the identifiers
  cmaxItems[maxCount] = 0;
	while ((ka < cntLine.length()) && (cntLine.charAt(ka) == ' ')) ++ka;
  pos = cntLine.indexOf(" ",ka);
  while (pos > 0) {
	 	subst = cntLine.substring(ka,pos).trim();
//		System.out.println(ka + "  " + pos + "  " + subst);  // debug
  	if (subst.indexOf('*') > 0) {
  		count[maxCount][cmaxItems[maxCount]] = subst.substring(0,subst.indexOf('*'));
  		cWildcard[maxCount][cmaxItems[maxCount]] = true;
  	}
  	else {
  		count[maxCount][cmaxItems[maxCount]] = subst;
  		cWildcard[maxCount][cmaxItems[maxCount]] = false;
  	}
  	++cmaxItems[maxCount];
  	ka = pos + 1;
  	while ((ka < cntLine.length()) && (cntLine.charAt(ka) == ' ')) ++ka;
		if (ka < cntLine.length()) pos = cntLine.indexOf(" ",ka);
  	else  pos = -1;
  }
  if (cLabel[maxCount].equals(""))   // create default label is label hasn't been set
  {
  	if (cWildcard[maxCount][0]) cLabel[maxCount] = count[maxCount][0] + "*";
  	else cLabel[maxCount] = count[maxCount][0];
  }
  this.maxCount++;

/* 	if (false) { // debugging: show the results
		System.out.println("Input: " + cntLine);
		System.out.println("Label: " + cLabel[maxCount - 1]);
		System.out.print("Events: ");
	  for (ka = 0; ka < cmaxItems[maxCount - 1]; ++ka) {
			System.out.print(count[maxCount - 1][ka]);
			if (cWildcard[maxCount - 1][ka]) System.out.print("*   ");
			else System.out.print("-  ");
		}
		System.out.println();
		System.out.println();
	}// debugging
*/
} // parse_count

void get_identifiers(String st, int index)
// extract identifiers from subst, put information in index Target element
{
int ka = 0;
int kb = 0;
String ident;

	dTmaxItems[index] = 0;
  dLabel[index] = "";
	st = st.concat(" ");
	kb = st.indexOf(' ',ka);
	while  (kb > 0) {
		ident = st.substring(ka,kb);
//		System.out.println(ident);
		ident = ident.trim();
		if (ident.indexOf('*') > 0)  // technically this should only be found at ident.length()-1
		{
	  	dTarget[index][dTmaxItems[index]] = ident.substring(0,ident.length()-1);
      dTWildcard[index][dTmaxItems[index]] = true;		
		}
		else 
		{
	  	dTarget[index][dTmaxItems[index]] = ident;
      dTWildcard[index][dTmaxItems[index]] = false;		
		}
		++dTmaxItems[index]; 
		ka = kb + 1;
		kb = st.indexOf(' ',ka);
	}	
} // get_identifiers

void process_identifiers(String subst, boolean isTarget)
// extract identifiers from a block in subst; set wildcard booleans
{
	if (this.maxTargets == TARGETSMAX) {
	  this.KEDS_error("Too many entries in Target section " + subst + " Ignored");
	  return;
 	}
	if (isTarget) {
		if (subst.indexOf("***") >= 0) // check for all-targets code 
		{ 
			tAll[maxTargets] = true; 
			dTmaxItems[maxTargets] = 0;
			dTarget[maxTargets][0] = "***";
		}
		else
		{
			tAll[maxTargets] = false;
			this.get_identifiers(subst,maxTargets);
		}
		++maxTargets;
	}
	else
	{
		this.get_identifiers(subst,0); // use the 0 element of dTarget as temp storage
		maxSItems = dTmaxItems[0];
		for (int kc = 0; kc < maxSItems; ++kc) {
			dSource[kc] = dTarget[0][kc];
			dWildcard[kc] = dTWildcard[0][kc];		
		}		 							  	
	}
} // process_identifiers

void parse_dyad(String dyadLine)
// extract identifiers and labels from the dyad line
{
boolean isTarget = false;
int ka = 0;
int kb = 0;
String subst;
StringBuffer ident = new StringBuffer("");
 
 maxTargets = 0;
 dyadLine = dyadLine.trim();
 dyadLine += ' ';
 do  
 {
	switch(dyadLine.charAt(ka)) {
		case ' ': 	++ka; break;
		case '/': 	isTarget = true;
								 ++ka;
								 break;
		case '{': 	if (!isTarget) this.Fatal_error("Error in " + dyadLine + ": {...} label blocks only apply to targets");
								kb = dyadLine.indexOf('}',ka);  // process a label
								if (kb>0) {
									subst = dyadLine.substring(ka+1,kb);
									dLabel[maxTargets-1] = subst.trim();
									ka = kb + 1; 		 							
								}
								else this.Fatal_error("Error in " + dyadLine + ": unmatched \'}\' in the substring " + dyadLine.substring(ka));
								break;
		case '[': 	kb = dyadLine.indexOf(']',ka);  // process an identifier block
								if (kb>0) {
									this.process_identifiers(dyadLine.substring(ka+1,kb), isTarget);
									ka = kb + 1; 		 							
								}
								else 
								{this.Fatal_error("Error in " + dyadLine + ": unmatched \']\' in the substring " + dyadLine.substring(ka));};
								break;
		 case ']': 
		 case '}':  this.Fatal_error("Error in " + dyadLine + ": unexpected \']\' or \'}\'");
								break;
		 default: 	ident.delete(0,ident.length()); // extract isolated identifier		 					
		 						while ((dyadLine.charAt(ka) != ' ')  && (dyadLine.charAt(ka) != '/')  && (dyadLine.charAt(ka) != '{')) { 
									ident.append(dyadLine.charAt(ka));
									++ka;
								}
								this.process_identifiers(ident.toString(), isTarget);
		}
//		System.out.print(dyadLine.charAt(ka) + " " + ka + " || ");
	} while (ka < dyadLine.length());
	
	for (ka = 0; ka < maxTargets; ++ka) {
		if (dLabel[ka].length() == 0) dLabel[ka] = this.dSource[0] + "2" + this.dTarget[ka][0];  // set default label
 	}
 	
/* 	if (false) { // debugging: show the results
		System.out.println("Input: " + dyadLine);
		System.out.print("Source: ");
		for (ka = 0; ka < maxSItems; ++ka) {
			System.out.print(dSource[ka]);
			if (dWildcard[ka]) System.out.print("*   ");
			else System.out.print("-  ");
		}
		System.out.println();
		System.out.println("\nTargets: ");
		for (kb = 0; kb < maxTargets; ++kb) {
			System.out.println(dLabel[kb]);
			for (ka = 0; ka < dTmaxItems[kb]; ++ka) {
				System.out.print(dTarget[kb][ka]);
				if (dTWildcard[kb][ka]) System.out.print("*   ");
				else System.out.print("-  ");
			}
		  System.out.println();
		} 
		System.out.println();
	}// debugging
*/
} // parse_dyad


void process_dyad()
{
String dataLine = null;
 
//  this.KEDS_error("KEDS_count processing dyad"  +	" status mark every " + this.formatMark + " entries");
  if (doDyadFiles) this.open_dyad_files();
  if (doComposFile) this.open_composite_file();

  this.firstData = true;  //  Needed to set begin Date
  if (maxSItems == 1) {this.dataDisplay.append("\n" + "Processing Source " + this.dSource[0] + "\n");}
  else  {
  	this.dataDisplay.append("\n" + "Processing Sources ");
  	for (int ka = 0; ka < maxSItems; ++ka) { this.dataDisplay.append(this.dSource[ka] + " ");}
  	this.dataDisplay.append("\n");
  }
  this.dspcnt = 0;
  this.dspline = 0;
  this.dspchar = " ";
  for ( int fileloop = 0; fileloop < this.maxFiles; fileloop++)
  {  // go thru the data files  this.files[]
   try
		{
			File dataFile = new File(this.files[fileloop]);
			FileInputStream datain = new FileInputStream(dataFile);
			BufferedReader br = new BufferedReader(new InputStreamReader(datain));
			while ((dataLine = br.readLine()) != null) { //  process the command file line by line
				if (this.parse_event(dataLine)) this.match_event();
			}  // while 
			datain.close();
  	}  // try
		catch(java.io.IOException e) { this.KEDS_error("Cannot read from data file" + this.files[fileloop]); }  
  } // file loop

// output final counts
	if (doDyadFiles)  {
		this.outputCurrDate();
		this.close_dyad_files();
	}
	if (doComposFile) {
		this.xferCurrDate();
		this.close_composite_file();
	}
}  // process_dyad

void open_dyad_files()
{
  String filename;
	
  for ( int loopvar = 0; loopvar < this.maxTargets; loopvar++) {
		filename = this.formatPrefix + dLabel[loopvar];
		try {
			dyadOut[loopvar] = new PrintWriter(new FileOutputStream(filename));
			dyadOut[loopvar].print("Date" + TAB + "N" + TAB + "Scale");
			for ( int loop2 = 0; loop2 < this.maxCount; loop2++) {
				dyadOut[loopvar].print(TAB + this.cLabel[loop2]);
			} // loop2
			this.dyadOut[loopvar].println();			
		}
		catch(java.io.IOException e2) {
			this.KEDS_error("Problem opening " + filename );
			return;
		}	
  } // for
} // open_dyad_files

void open_composite_file()
// open composite and temp files
{
  StringBuffer scount = new StringBuffer (""); 
  
	try { tempOut = new PrintWriter(new FileOutputStream(TEMPFILENAME)); }
	catch(java.io.IOException e2) {
		this.KEDS_error("Problem opening temporary output file" + TEMPFILENAME );
		return;
	}
	if (this.hasCompos)
	{
		try {
				this.composIn = new BufferedReader(new FileReader(this.composFileName));
				scount.append(composIn.readLine()); // read column label line
		} 
		catch (IOException e) {
			this.KEDS_error("Problem opening .count output file" + this.composFileName);
			return;
		}
	}
	else this.composFileName = this.formatPrefix + "counts";
	
  for ( int loopvar = 0; loopvar < this.maxTargets; loopvar++)
  {
		if (scount.length() > 0) scount.append(TAB);
		scount.append(dLabel[loopvar] + TAB + "N" + TAB + "Scale");
		for ( int loop2 = 0; loop2 < this.maxCount; loop2++) {
			scount.append(TAB + this.cLabel[loop2]);
		} // loop2
  } // for
  this.tempOut.println(scount);

} // open_composite_file

void close_dyad_files()
{
	for ( int loopvar = 0; loopvar < this.maxTargets; loopvar++)
	  { this.dyadOut[loopvar].close(); } 
} // close_dyad_files

void close_composite_file()
// close and rename temp file
{
	this.tempOut.close(); 
  if (this.hasCompos) {
		try {this.composIn.close();}
		catch(java.io.IOException e2) { 
			this.KEDS_error("Problem closing " + this.composFileName );
			return;
		}
	}
  else { this.hasCompos = true; }

// delete composite file if it exists, then rename the temp file
	boolean exists = (new File(this.composFileName)).exists();  // note that a file by this name could have been around from earlier run
	if (exists) {
		boolean success = (new File(this.composFileName)).delete();
		if (!success) this.KEDS_error("Problem deleting " + this.composFileName + " file");  // may want to do something here someday...
	}

	File file = new File(TEMPFILENAME);  // ### make this a constant
  File file2 = new File(this.composFileName);
	boolean success = file.renameTo(file2);
	if (!success) this.KEDS_error("Problem renaming " + TEMPFILENAME + " file"); 

} // close_composite_files

void match_event()
// core procedure: check whether the event meets the criteria, and do the
// appropriate counts
{
boolean match = false;
int			ka = 0;

  if (CurrDate.before(FirstDate)) return; // skip data before FirstDate
  
  // update the display
  if (this.dspline >= 45) { // ### magic number here...
	  this.dataDisplay.append("\n");
	  this.dspline = 0;
	}
  if (this.dspcnt >= this.formatMark ) {
	  this.dataDisplay.append(this.dspchar);
	  this.dspcnt = 0;
	  this.dspline++;
	  this.dspchar = " ";
	}

	// evaluate source condition
	ka = 0;
	while (!match && ka < maxSItems) {
		if ((dSource[ka].equals(locSrc)) ||
				(dWildcard[ka] && locSrc.startsWith(dSource[ka])))
		{ match = true; }
		++ka;
	} 

	if (!match) { 
		if (this.dspchar.equals(" ")) { this.dspchar = "-"; }
	  this.dspcnt++;
		return;
	}
	
	// evaluate target conditions
	for (int loop = 0; loop < maxTargets; loop++)
 	{
		match = tAll[loop];  // handle the '***' all-events condition
		ka = 0;
		while (!match && ka < dTmaxItems[loop]) {
			if ((dTarget[loop][ka].equals(locTar)) ||
				(dTWildcard[loop][ka] && locTar.startsWith(dTarget[loop][ka])))
			{ match = true; }
			++ka;
		} 
		if (!match) //  adjust the display then try next target set
		{  
			if (this.dspchar.equals(" ") || this.dspchar.equals("-")) {  this.dspchar = "*"; } 
			this.dspcnt++; 
			continue;
		}

		// adjust the display -- probably isn't handling the multiple options correctly
		if (this.dspchar.equals(" ") || this.dspchar.equals("-") ||	this.dspchar.equals("*"))
		{  this.dspchar = "+"; }
		this.dspcnt++;

		this.dayCount[loop]++;
		for (int sclLoop = 0; sclLoop < maxCodeValue; sclLoop++) // add scale value  ### this should be a while loop!
		{  if (locEvt.equals(code[sclLoop]))
			 {this.dayScale[loop] += this.value[sclLoop]; }
		}

		for (int cntLoop = 0; cntLoop < maxCount; cntLoop++) {  
			ka = 0;
			match = false;
			while (!match && ka < cmaxItems[cntLoop]) {
				if ((count[cntLoop][ka].equals(locEvt)) ||
				   (cWildcard[cntLoop][ka] && this.locEvt.startsWith(count[cntLoop][ka]))) 
				{ 
					++this.dayCode[loop][cntLoop]; 
					match = true;
				}
				++ka;
			} // while
		} // cntLoop

	}// for loop thru targets
} // match_event


boolean parse_event(String dline)
{ // divide dataLine into Date, Src, Tar, Evt
  // based on columns in format

 int 		tabpos, column;
 String 	eventLine, eventItem;
 
 eventLine = dline;  // set the global variables to null values
 column = 0;
 locDate = null;
 locSrc = null;
 locTar = null;
 locEvt = null;
 
 while (eventLine != null)  // loop through the line by fields
 {
   tabpos = eventLine.indexOf(TAB);  // check for a tab
   if (tabpos <= 0)
   {
     eventItem = eventLine.trim();  // we're done
     eventLine = null;
   }
   else // put the next field into eventItem
   {
     eventItem = eventLine.substring(0,tabpos).trim();
     eventLine = eventLine.substring(tabpos+1);
   }
  column++;
  if (column == formatDate)  // assign to the appropriate variable
   	{
	  locDate = eventItem;
	  this.setCurrDate();
	}
   else if (column == formatSource) {
	  locSrc = eventItem;
 	}
   else if (column == formatTarget) {
	  locTar = eventItem;
	}
   else if (column ==formatEvent) {
	  locEvt = eventItem;
	}
 }  // while
 if ((locDate == null) || (locSrc == null) || (locTar == null)
			       || (locEvt == null))
   {
     this.KEDS_error("Bad data line");
     return false;
   } 
 else {return true;}
} // parse event

void setFirstDate()
// establish the initial point in the data set based on the first record read;
// this is called once; subsequently anything prior to FirstDate is skipped
{

	this.FirstDate.setTime(this.CurrDate.getTime()); 
	if (this.formatFreq.equals("MONTHLY") ) {  
		if (CurrDate.get(Calendar.DAY_OF_MONTH) != 1) {
			FirstDate.add(Calendar.MONTH,1); // start in next month
			FirstDate.set(Calendar.DAY_OF_MONTH,1);
		}
	}
	else if (this.formatFreq.equals("QUARTERLY") ) {  
		if (CurrDate.get(Calendar.DAY_OF_MONTH) != 1) {
			FirstDate.add(Calendar.MONTH,1); // start in next month
			FirstDate.set(Calendar.DAY_OF_MONTH,1);
		}
		while (FirstDate.get(Calendar.MONTH) % 3 > 0) { FirstDate.add(Calendar.MONTH,1);} // start on even quarter; JAN == 0 in Java
	}
	if (this.formatFreq.equals("WEEKLY") || this.formatFreq.equals("BIWEEKLY")) { 
		while (FirstDate.get(Calendar.DAY_OF_WEEK) != firstWeekDay) { FirstDate.add(Calendar.DATE,1);} // start on SUNDAY or MONDAY
	} 
	this.BeginDate.setTime(this.FirstDate.getTime()); 
  this.setEndDate();
	this.firstData = false;
} // getFirstDate()

void setCurrDate()
{ // take locDate and put into Curr Date
  // Assume date input in YYMMDD or YYYYMMDD format 
  // 1900 window 28 to 99

int cmonth = 0;
int cday   = 1;
int cyear  = 1999;
int posm = 0;
  
// convert date to Calendar format
if (this.locDate.length() == 6) {
    cyear = Integer.parseInt(this.locDate.substring(0,2));
    posm = 2;
    if (cyear >= 28) cyear += 1900; 
    else cyear += 2000;
  }
  else if (this.locDate.length() == 8) {
    cyear = Integer.parseInt(this.locDate.substring(0,4));
    posm = 4;
  }
  cmonth = Integer.parseInt(this.locDate.substring(posm,posm+2));
  cday = Integer.parseInt(this.locDate.substring(posm+2,posm+4));

	this.CurrDate.set(cyear,cmonth-1,cday);
//  this.printDate();  // debugging
  if (this.firstData) setFirstDate();  

  if (this.CurrDate.after(this.EndDate) || this.CurrDate.equals(this.EndDate)) { // output counts unitl date in range
  	do {
			if (doDyadFiles) this.outputCurrDate();
			if (doComposFile) this.xferCurrDate();
			this.resetCounts();
			this.BeginDate.setTime(this.EndDate.getTime()); 
			this.setEndDate();
		}
	 while (this.EndDate.before(this.CurrDate));
   }
} // setCurrDate

void setEndDate()
{ // set End Date to BeginDate + Frequency
  this.EndDate.setTime(this.BeginDate.getTime()); 
  if (this.formatFreq.equals("DAILY")) this.EndDate.add(Calendar.DATE,1);
  else if (this.formatFreq.equals("WEEKLY")) this.EndDate.add(Calendar.DATE,7);
  else if (this.formatFreq.equals("BIWEEKLY")) this.EndDate.add(Calendar.DATE,14);
  else if (this.formatFreq.equals("MONTHLY")) this.EndDate.add(Calendar.MONTH,1);
  else if (this.formatFreq.equals("QUARTERLY")) this.EndDate.add(Calendar.MONTH,3);
  else this.Fatal_error("Unkown Frequency ");  
} // setEndDate

/* void printDate()
// debugging output
{
	System.out.println( this.formatSDF.format( this.BeginDate.getTime() ));
	System.out.println( this.formatSDF.format( this.EndDate.getTime() ));
	System.out.println( this.formatSDF.format( this.CurrDate.getTime() ));
} // printDate
*/

void outputCurrDate()
{ // output current date info to output files
  for ( int loopvar = 0; loopvar < this.maxTargets; loopvar++) {
		this.dyadOut[loopvar].print( this.formatSDF.format( this.BeginDate.getTime() ));
		this.dyadOut[loopvar].print(TAB + this.dayCount[loopvar] + TAB + this.dayScale[loopvar]);
    for ( int loop2 = 0; loop2 < this.maxCount; loop2++) {
			this.dyadOut[loopvar].print(TAB + this.dayCode[loopvar][loop2]);
		 
		} // loop2
    this.dyadOut[loopvar].println();
  } // for
} // outputCurrDate

void xferCurrDate()
{ // append current date info to composite file
  StringBuffer scount = new StringBuffer (""); 

  if (this.hasCompos)  { 
  	try {
  		scount.append(composIn.readLine()); 
//  	System.out.println("Fr count " + scount); // debug
  	} 
		catch(java.io.IOException e2) { 
			this.KEDS_error("Problem reading " + this.composFileName );
			return;
		}
	}
  for ( int loopvar = 0; loopvar < this.maxTargets; loopvar++) {
		if (scount.length()>0) {scount.append(TAB);}
		scount.append(this.formatSDF.format( this.BeginDate.getTime() ));
		scount.append(TAB + this.dayCount[loopvar] + TAB + this.dayScale[loopvar]);
    for ( int loop2 = 0; loop2 < this.maxCount; loop2++) {
			scount.append(TAB + this.dayCode[loopvar][loop2]);		 
		} // loop2
  } // for
	tempOut.println(scount);
//	System.out.println("To temp  " + scount);  // debug
} // xferCurrDate

void resetCounts()
{ // initialize counts to zero
  for ( int loopvar = 0; loopvar < this.maxTargets; loopvar++) {
		this.dayCount[loopvar] = 0;
		this.dayScale[loopvar] = 0;
		for ( int loopcnt = 0; loopcnt < this.maxCount; loopcnt ++)
    { this.dayCode[loopvar][loopcnt] = 0;}
  } // for
} // resetCounts


 public static void main( String args[]) 
{ 
  KEDS_count 	kc_dsp;
  Button 			kc_quit;
  TextArea 		kc_title;
  
  // Initializations
  kc_dsp 		        = new KEDS_count();

  kc_dsp.code       = new String[CODEVALUEMAX];
  kc_dsp.value      = new float[CODEVALUEMAX];
  kc_dsp.files   	  = new String[FILESMAX];

  kc_dsp.count    	= new String[COUNTMAX][ITEMSMAX];
  kc_dsp.cWildcard  = new boolean[COUNTMAX][ITEMSMAX];
  kc_dsp.cmaxItems 	= new int[TARGETSMAX];
  kc_dsp.cLabel     = new String[TARGETSMAX];
  kc_dsp.dSource    = new String[ITEMSMAX];
  kc_dsp.dWildcard  = new boolean[ITEMSMAX];
  kc_dsp.dTarget    = new String[TARGETSMAX][ITEMSMAX];
  kc_dsp.dTWildcard = new boolean[TARGETSMAX][ITEMSMAX];
  kc_dsp.dTmaxItems = new int[TARGETSMAX];
  kc_dsp.dLabel     = new String[TARGETSMAX];
  kc_dsp.tAll		    = new boolean[TARGETSMAX];

  kc_dsp.dyadOut    = new PrintWriter[TARGETSMAX];
  kc_dsp.dayCount	  = new int[TARGETSMAX];
  kc_dsp.dayScale	  = new float[TARGETSMAX];
  kc_dsp.dayCode	  = new int[TARGETSMAX][COUNTMAX];

  kc_dsp.BeginDate	= Calendar.getInstance();
  kc_dsp.EndDate	  = Calendar.getInstance();
  kc_dsp.CurrDate	  = Calendar.getInstance();
  kc_dsp.FirstDate	= Calendar.getInstance();
  
  kc_dsp.formatSDF = new SimpleDateFormat("yyMMdd");


  // Set up program frame. 
  kc_dsp.setTitle("KEDS count");
  kc_dsp.addWindowListener(new WindowAdapter()
     { public void windowClosing(WindowEvent e) { System.exit(0);} }  
  );  
  kc_title  = new TextArea("",3,50,TextArea.SCROLLBARS_NONE);
  kc_title.append("KEDS_count: Event data aggregation program \n");
  kc_title.append("Kansas Event Data System " + VER + "\n");
//  kc_title.append(" 1995, Philip A. Schrodt");
  kc_quit   = new Button("Quit and Exit");
  kc_quit.addActionListener(kc_dsp);
  kc_dsp.status = new TextField("STATUS: KEDS_count starting");
  kc_dsp.dataDisplay = new TextArea("",36,80,TextArea.SCROLLBARS_BOTH);
  kc_dsp.add(kc_title,BorderLayout.NORTH);
  kc_dsp.add(kc_quit,BorderLayout.EAST);
  kc_dsp.add(kc_dsp.status,BorderLayout.WEST);  
  kc_dsp.add(kc_dsp.dataDisplay,BorderLayout.SOUTH);
  kc_dsp.pack();
  kc_dsp.setVisible(true);

  // Process data
  kc_dsp.status.setText("STATUS: Reading in command file");

  kc_dsp.KEDS_read_command_file();
	
	kc_dsp.status.setText("STATUS: Closing Files");
  System.out.println("KEDS_count ended - Press Quit" + "\007");
}  //  End of main


}  //  End of KEDS_count class 
